package com.wombat.aliyun_oss

import android.app.Activity
import android.app.Activity.RESULT_OK
import android.content.Intent
import android.net.Uri
import android.os.Environment
import android.os.Handler
import android.os.Looper
import android.util.Log
import android.webkit.MimeTypeMap
import androidx.annotation.NonNull
import com.alibaba.sdk.android.oss.ClientConfiguration
import com.alibaba.sdk.android.oss.OSSClient
import com.alibaba.sdk.android.oss.common.auth.OSSPlainTextAKSKCredentialProvider
import com.alibaba.sdk.android.oss.common.auth.OSSStsTokenCredentialProvider
import com.alibaba.sdk.android.oss.common.utils.BinaryUtil
import com.wombat.aliyun_oss.listener.DownloadCallback
import com.wombat.aliyun_oss.listener.UploadCallback
import com.wombat.aliyun_oss.model.*
import com.wombat.aliyun_oss.model.download.*
import com.wombat.aliyun_oss.model.upload.*
import com.wombat.aliyun_oss.model.upload.UploadSuccessData
import com.wombat.aliyun_oss.service.OSSService
import com.wombat.aliyun_oss.utils.FileUtil
import com.wombat.aliyun_oss.utils.MapUtil
import com.wombat.aliyun_oss.utils.RealPathUtil
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.EventChannel
import io.flutter.plugin.common.EventChannel.EventSink
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry
import java.io.File
import java.util.*
import kotlin.collections.HashMap
import kotlin.concurrent.thread

/** AliyunOSSPlugin */
class AliyunOSSPlugin : FlutterPlugin, MethodCallHandler, ActivityAware,
    PluginRegistry.ActivityResultListener, EventChannel.StreamHandler {
    private lateinit var methodChannel: MethodChannel
    private var activity: Activity? = null
    private var ossService: OSSService? = null
    private var credentialData: CredentialData? = null
    private var activityPluginBinding: ActivityPluginBinding? = null
    private var bucketName: String = ""

    private val resultSelectFile = 101

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        methodChannel =
            MethodChannel(flutterPluginBinding.binaryMessenger, "com.wombat/aliyun_oss_plugin")

        methodChannel.setMethodCallHandler(this)

    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        methodChannel.setMethodCallHandler(null)
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        activity = binding.activity
        binding.addActivityResultListener(this)
        activityPluginBinding = binding
    }

    override fun onDetachedFromActivityForConfigChanges() {

    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {

    }

    override fun onDetachedFromActivity() {
        activity = null
        activityPluginBinding?.removeActivityResultListener(this)
        activityPluginBinding = null
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?): Boolean {
        // 用于处理选择文件的结果监听
        if (resultCode != RESULT_OK) {
            return false
        }
        if (requestCode == resultSelectFile && null != data) {
            val filePath = RealPathUtil.getRealPath(activity!!, data.data!!)
            if (filePath != null) {
                bucketName = "wombat-oss1"
                upload(bucketName, data.data!!, filePath, null, UUID.randomUUID().toString())
                return true
            }
        }
        return false
    }

    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        when (call.method) {
            "init" -> {
                val initResult = initOSSClient(call.arguments as HashMap<String, Any>)
                result.success(mapOf("initResult" to initResult))
            }

            "saveFileToDownload" -> {
                val fileName = call.argument<String>("fileName") as String
                val mimeType = call.argument<String>("mimeType") as String
                val path = saveFileToDownload(fileName, mimeType)
                Log.d("保存文件地址", "$path")
                result.success(path)
            }

            "getMd5Str" -> {
                val filePath = call.argument<String>("filePath") as String
                thread {
                    val md5 = BinaryUtil.calculateMd5Str(filePath)
                    activity?.runOnUiThread {
                        result.success(md5)
                    }
                }
            }

            "uploadWithPath" -> {
                val bucketName: String = call.argument<String>("bucketName") as String
                val filePath: String = call.argument<String>("filePath") as String
                val md5: String = call.argument<String>("md5") as String
                val uuid: String = call.argument<String>("uuid") as String
                upload(bucketName, Uri.fromFile(File(filePath)), filePath, md5, uuid)
            }

            "upload" -> {
                selectFile(call.argument<String>("bucketName") as String)
            }

            "pause_upload" -> {
                val objectKey = call.argument<String>("objectKey") as String
                val uuid = call.argument<String>("uuid") as String
                val pauseResult = ossService?.pauseUpload(objectKey, uuid)
                result.success(pauseResult)
            }

            "resume_upload" -> {
                val objectKey = call.argument<String>("objectKey") as String
                val uuid = call.argument<String>("uuid") as String
                val resumeResult = ossService?.resumeUpload(objectKey, uuid)
                result.success(resumeResult)
            }

            "cancel_upload" -> {
                val objectKey = call.argument<String>("objectKey") as String
                val uuid = call.argument<String>("uuid") as String
                ossService?.cancelUpload(objectKey, uuid)
                result.success(objectKey)
            }

            "cancel_all_upload" -> {
                ossService?.cancelAllUpload()
                result.success(bucketName)
            }

            "download" -> {
                val tempBucketName: String? = call.argument<String>("bucketName") as String?
                val md5: String = call.argument<String>("md5") as String
                val fileName: String = call.argument<String>("fileName") as String
                val mimeType: String = call.argument<String>("mimeType") as String
                download(tempBucketName ?: bucketName, md5, fileName, mimeType)
            }

            "pause_download" -> {
                val objectKey = call.argument<String>("objectKey") as String
                val pauseResult = ossService?.pauseDownload(objectKey)
                result.success(pauseResult)
            }

            "resume_download" -> {
                val objectKey = call.argument<String>("objectKey") as String
                val resumeResult = ossService?.resumeDownload(objectKey)
                result.success(resumeResult)
            }

            else -> result.notImplemented()
        }
    }

    override fun onListen(arguments: Any?, events: EventSink?) {
        startListening(arguments, events)
    }

    override fun onCancel(arguments: Any?) {
        cancelListening(arguments)
    }

    private var listeners = HashMap<Any?, Runnable>()

    private fun startListening(listener: Any?, emitter: EventSink?) {
        val handler = Handler(Looper.getMainLooper())
        listeners[listener] = object : Runnable {
            override fun run() {
                if (listeners.containsKey(listener)) {
                    // Send some value to callback
                    emitter?.success("Hello listener! " + System.currentTimeMillis() / 1000)
                    handler.postDelayed(this, 1000)
                }
            }
        }
        // Run task
        handler.postDelayed(listeners[listener]!!, 1000)
    }

    private fun cancelListening(listener: Any?) {
        listeners.remove(listener);
    }

    /**
     * 初始化 OSS 客户端
     */
    private fun initOSSClient(params: HashMap<String, Any>): Boolean {

        val endpoint = params["endpoint"] as String
        val accessKeyId = params["accessKeyId"] as String
        val secretKeyId = params["secretKeyId"] as String
        val securityToken = params["securityToken"] as String?

        val newCredentialData = CredentialData(
            endpoint = endpoint,
            accessKeyId = accessKeyId,
            secretKeyId = secretKeyId,
            securityToken = securityToken
        )

        /// 值相同，则不重新初始化oss_client
        if (credentialData == newCredentialData) {
            return false
        }

        credentialData = newCredentialData

        val conf = ClientConfiguration().apply {
            connectionTimeout = 15 * 1000 // 连接超时，默认15秒

            socketTimeout = 15 * 1000 // socket超时，默认15秒

            maxConcurrentRequest = 5 // 最大并发请求数，默认5个

            maxErrorRetry = 2 // 失败后最大重试次数，默认2次
        }

        val credential = if (securityToken == null) {
            OSSPlainTextAKSKCredentialProvider(
                accessKeyId,
                secretKeyId
            )
        } else {
            OSSStsTokenCredentialProvider(
                accessKeyId,
                secretKeyId,
                securityToken
            )
        }
        ossService = OSSService.getInstance(
            oss = OSSClient(
                activity!!,
                endpoint,
                credential,
                conf
            ),

            recordDirectory = activity!!.filesDir.absolutePath.toString() + "/oatos/oss_record/",
            uploadCallback = UploadCallback(
                this::uploadInitCallback,
                this::uploadProgressCallback,
                this::uploadSuccessCallback,
                this::uploadFailCallback
            ),
            downloadCallback = DownloadCallback(
                this::downloadInitCallback,
                this::downloadProgressCallback,
                this::downloadSuccessCallback,
                this::downloadFailCallback
            ),
            context = activity!!.applicationContext
        )

        return true
    }

    private fun saveFileToDownload(fileName: String, mimeType: String): String? {
        val path =
            Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS)?.path
        return path?.run {
            FileUtil.getUniqueFileName(this, fileName).path
        }
    }

    private fun selectFile(bucketName: String) {
        this.bucketName = bucketName
        activity?.run {
            val intent = Intent(
                Intent.ACTION_GET_CONTENT
            )
            intent.type = "*/*"
            startActivityForResult(intent, resultSelectFile)
        }
    }

    /**
     * 上传文件，回传上传进度，处理分片/断点续传
     */
    private fun upload(
        bucketName: String,
        fileUri: Uri,
        filePath: String,
        inputMd5: String? = null,
        uuid: String
    ) {
        val extension =
            if (filePath.contains(".")) {
                filePath.substring(filePath.lastIndexOf("."))
            } else {
                ""
            }

        thread {
            val md5: String = if (inputMd5 == null) {
                val start = System.currentTimeMillis()
                Log.d("计算md5", filePath)
                val result = BinaryUtil.calculateMd5Str(filePath)
                Log.d("計算md5耗時", "${System.currentTimeMillis() - start}")
                result
            } else {
                inputMd5
            }
            val objectKey = "$md5$extension"

            Log.d("上传名称objectKey", objectKey)

            uploadInitCallback(
                UploadInitData(
                    filePath = filePath,
                    objectKey = objectKey,
                    uuid = uuid
                )
            )
            val extensionWithoutDot =
                if (extension.isNotEmpty()) extension.substring(1) else extension
            val mimeType = MimeTypeMap.getSingleton().getMimeTypeFromExtension(extensionWithoutDot)
            Log.i("mimeType类型", mimeType + "")
            ossService?.upload(
                bucketName,
                objectKey,
                fileUri,
                uuid,
                mimeType
            )

        }
    }

    /**
     * 下载文件，回传下载进度，处理分片/断点续传
     */
    private fun download(
        bucketName: String,
        md5: String,
        fileName: String,
        mimeType: String,
    ) {
        val extension =
            if (fileName.contains(".")) {
                fileName.substring(fileName.lastIndexOf("."))
            } else {
                fileName
            }

        val objectKey = "$md5$extension"

        Log.d("下载名称objectKey", objectKey)
        val filePath = saveFileToDownload(fileName, mimeType)
        if (filePath == null || filePath == "") {
            Log.d("下载文件初始失败", "文件路径为空")
            return
        }
        val fileUri = Uri.fromFile(File(filePath))
        downloadInitCallback(DownloadInitData(filePath = filePath, objectKey = objectKey))

        ossService?.download(
            bucketName,
            objectKey,
            fileUri
        )
    }

    private fun runOnUiThread(callback: () -> Unit) {
        activity?.runOnUiThread {
            callback()
        }
    }

    private fun uploadInitCallback(initData: UploadInitData) {
        runOnUiThread {
            methodChannel.invokeMethod("upload_init", MapUtil.toMap(initData))
        }
    }

    private fun uploadProgressCallback(
        progress: UploadProgressData
    ) {
        runOnUiThread {
            methodChannel.invokeMethod(
                "upload_progress",
                MapUtil.toMap(progress)
            )
        }
    }

    private fun uploadSuccessCallback(successData: UploadSuccessData) {
        runOnUiThread {
            methodChannel.invokeMethod(
                "upload_success",
                MapUtil.toMap(successData)
            )
        }
    }

    private fun uploadFailCallback(failData: UploadFailData) {
        runOnUiThread {
            methodChannel.invokeMethod(
                "upload_fail",
                MapUtil.toMap(failData)
            )
        }
    }

    private fun downloadInitCallback(initData: DownloadInitData) {
        runOnUiThread {
            methodChannel.invokeMethod("download_init", MapUtil.toMap(initData))
        }
    }

    private fun downloadProgressCallback(
        progress: DownloadProgressData
    ) {
        runOnUiThread {
            methodChannel.invokeMethod(
                "download_progress",
                MapUtil.toMap(progress)
            )
        }
    }

    private fun downloadSuccessCallback(successData: DownloadSuccessData) {
        runOnUiThread {
            methodChannel.invokeMethod(
                "download_success",
                MapUtil.toMap(successData)
            )
        }
    }

    private fun downloadFailCallback(failData: DownloadFailData) {
        runOnUiThread {
            methodChannel.invokeMethod(
                "download_fail",
                MapUtil.toMap(failData)
            )
        }
    }
}
